#!/usr/bin/env perl
# Last changed Time-stamp: <2010-10-15 14:43:08 joe>
# joe@bioinf.uni-leipzig.de
########################################################################

use strict;
use warnings;
use SynBlast::MyUtils;

package SynBlast::MySyntenyUtils; 

#require Exporter;
#my @ISA = qw(Exporter);
#my @EXPORT = qw(blastfile2Lists infofile2Exonlist info2Infofile blastfile2sortedLists getSynRegInfo);
#my @EXPORT_OK = @EXPORT;


################# moved from doSyntenyFilter.pl ###########3

sub createSyntenyRegions {
		my($blastHits, $minQN, $flP, $realSlength, $mScThreshold, $flSize, $stepGrain, $gFl 
		, $writeFlag, $DBname, $synResultDir, $cmdlineWas, $EnsemblD, $restrChr, $maxPercOverlap, $nFl, $resRegTable, $silentFl) 
		= SynBlast::MyUtils::rearrange( [qw(
									BLASTHIT_HASH
									MIN_QN
									FLANKING_FAC
									REAL_SIZE
									MIN_SCORE_CUTOFF
									FLANKING_SIZE
									STEP_GRAIN
									GLOBAL_MAX
									SAVE_FLAG
									DB_NAME
									RESULT_DIR
									COMMAND_STR
									ENSEMBL_DATE
									RESTRICT_CHR
									MAX_PERC_OVL
									CHR_MAX
									RESULT_TABLE
									SILENT
									)], @_);
									
	%{$resRegTable}=() unless(defined($resRegTable));
	
	
	

	## start processing of chromosomes in order of total hits per chromosome
	my %blHitChrCounts=();
	foreach  my $cchr (keys %{$blastHits}) {
		push(@{$blHitChrCounts{keys(%{$blastHits->{$cchr}})}}, $cchr);
	}
	my %allRegionsBySc=();
	#my $savedFiles="";
	my @savedFiles=();
	foreach  my $cnt (sort {$a <=> $b} (keys %blHitChrCounts)) {	
		if($cnt < $minQN) {
			print "." x scalar(@{$blHitChrCounts{$cnt}}) unless($silentFl); 
			next;
		}
		foreach  my $cchr (@{$blHitChrCounts{$cnt}}) {	
			
			next if($restrChr && ($restrChr ne $cchr));
			
						
			#calculate SlidingWindow-value to identify regions
			my $distribution=SynBlast::MySyntenyUtils::calcSlidingWindowStat($blastHits->{$cchr},  $flP,  $realSlength, $minQN, $mScThreshold, $flSize, $stepGrain, $silentFl);
			next unless(defined($distribution));
			
			
			
			my %regionsByMaxSc=();
			
						
			foreach my $ii (keys %{$distribution}) {
				
			
					
				my $percOfMaxRange=int(($distribution->{$ii}->[3] - $distribution->{$ii}->[6] +1)/($flSize) * 1000)/10.;

				### now use as second order-by field/criteria the total sum of scores (for regions of same max-score-sum-per-query)
				 ### 																* indicates where information is w.r.t. only those hits in a region, 
				 ###																	that are above a scoreThreshold ($mScThreshold)
				 ### 					SumOf_MaxScorePerQuery*,   totalScoreSum*
			### 					diffQN   SumOf_MaxScorePerQuery*,   totalScoreSum*
				#push(@{$regionsByMaxSc{$distribution->{$ii}->[2]}->{$distribution->{$ii}->[4]}->{$distribution->{$ii}->[1]}},  [ (
				push(@{$regionsByMaxSc{"1"}->{$distribution->{$ii}->[4]}->{$distribution->{$ii}->[1]}},  [ (
									$ii  											#  genom. start (of sl.window) 
									,  $distribution->{$ii}->[6]  						# genomic start used, 
									,  $distribution->{$ii}->[3]						# genomic end used,  
									, $distribution->{$ii}->[2]						#number of diff. queries *
									,  $distribution->{$ii}->[0]						#total number of HSPs contained 
									, $distribution->{$ii}->[1]  						## total score sum *
									,  $percOfMaxRange							### percent of maxRangeUsed
									) ]);
			}
			#next;
			my $outstr="";
			my $breakFl=0;
			my @doneIntervs=();
			## sort order of evaluation is decreasing, first in SumOfMaxScoresPerQuery, second by totalScoreSum
			
			foreach my $QNkey (reverse (sort {$a <=> $b} (keys %regionsByMaxSc) ) ) {
				foreach my $scKey (reverse (sort {$a <=> $b} (keys %{$regionsByMaxSc{$QNkey}} ) ) ) {
					foreach my $tscKey (reverse (sort {$a <=> $b} (keys %{$regionsByMaxSc{$QNkey}->{$scKey}} ) ) ) {
						foreach my $entr  (@{$regionsByMaxSc{$QNkey}->{$scKey}->{$tscKey}}) {
							next if($entr->[3] < $minQN); ## QN not reached?
						
						### check on overlap with previous interval set
						my $oovlp=0;
						foreach my $intvl (@doneIntervs) {
							my $ooolp= (SynBlast::MyUtils::getOverlapSize(@{$intvl},  $entr->[1], $entr->[2]));
							$oovlp +=1 if($ooolp > int($maxPercOverlap * (abs($entr->[2]-$entr->[1]) + 1) / 100. )); 
							### only skip overlapping candidates if overlapping more than maxPercOverlap% in length
						}
						next if($oovlp);	#print "/";
							
						my $regNaame=$scKey . "\t" . join("\t", @{$entr});
						#$outstr .= $regNaame;
						$outstr .= #$QNkey . "\t" . 
								$regNaame;
							
						if((@doneIntervs < $nFl)) {  ### save region to syntenyRegion-File...
							if(! $gFl) {
								if($writeFlag) {
									$outstr .= "\t(saved)\n";
									# write region to file
									
									#$savedFiles .= SynBlast::MySyntenyUtils::writeSynRegToFile(
									 my $csFile=SynBlast::MySyntenyUtils::writeSynRegToFile(
										'-REG_NR' => @doneIntervs +1
										, '-REG_NAME' => $regNaame
										, '-REG_HITS' => $distribution->{$entr->[0]}->[5]
										, '-ENTRY' => $entr
										, '-CHR' => $cchr
										, '-REAL_SIZE' => $realSlength
										, '-FLANKING_FAC' => $flP
										#, '-GREG_NR' =>  undef
										, '-FLANKING_SIZE' => $flSize
										, '-DB_NAME' =>$DBname
										, '-MIN_QN' => $minQN
										, '-RESULT_DIR' => $synResultDir
										, '-COMMAND_STR' => $cmdlineWas
										, '-ENSEMBL_DATE'=> $EnsemblD
										);
									#$savedFiles .=$csFile;
									push(@savedFiles, $csFile);
																		
								} else {
									$outstr .= "\t(s)\n";	
								}
							} else {  ## only mark for writing/saving (global Max Number set)
								if(@doneIntervs < $gFl) {
									$outstr .= "\t(marked for saving)\n";
									
									push(@{$allRegionsBySc{$QNkey}->{$scKey}},  [ ($regNaame
													, $entr
													, $cchr
													, $distribution->{$entr->[0]}->[5]
													, @doneIntervs +1
													) ] );
								} else {
									$outstr .= "\n";	
								}
							}
						} else {
							$outstr .= "\n";	
						}
					
						push(@doneIntervs, [ ($entr->[1], $entr->[2]) ] );
						if((! $writeFlag) && (@doneIntervs >= $nFl)) {
							++$breakFl;
							last;
						}
					}
					last if($breakFl);
				}
				last if($breakFl);
			  }
			  last if($breakFl);
		        }
			if(@doneIntervs) {
				print "chromosome ${cchr}:\n" . $outstr  unless($silentFl);
			} else {
				$|=1;
				print "." unless($silentFl);
			}
			## go to next chr
		}
       }
       	print "files summary:\n" . join("", @savedFiles) if(@savedFiles && (! $silentFl));
			       
       if($gFl) { ## write out globally max gFl best hits (by score)  ## have to add ordering by QN first!!!
		my $breakFl=0;
		my $doneRegs=0;
		my $outstr="";
		#my $savedFiles="";
		(! @savedFiles) || die "ERROR: savedFiles-Array should not contain any entries here!\n";
		
		$resRegTable->{'colnames'}= [ qw(SumOfMaxScores chr start rstart end QN totalHSPcount totalscore PercOfRange savedFlag) ];
		$resRegTable->{'data'}= [()];
	       #foreach my $scKey (reverse (sort {$a <=> $b} (keys %allRegionsBySc) ) ) {
	       foreach my $QNkey (reverse (sort {$a <=> $b} (keys %allRegionsBySc) ) ) {
			foreach my $scKey (reverse (sort {$a <=> $b} (keys %{$allRegionsBySc{$QNkey}}) ) ) {
			 foreach my $entr  (@{$allRegionsBySc{$QNkey}->{$scKey}}) {
				my $regNaame=$entr->[2] . "\t" . $entr->[0];
				$outstr .= $regNaame;
				push(@{$resRegTable->{'data'}},  [ ($scKey, $entr->[2], @{$entr->[1]}) ] );
				if($writeFlag && ($doneRegs < $gFl)) {  ### save region to syntenyRegion-File...
					$outstr .= "\t(saved)\n";
					# write region to file
					my $csFile= SynBlast::MySyntenyUtils::writeSynRegToFile(
										'-REG_NR' => $entr->[4]
										, '-REG_NAME' => $regNaame
										, '-REG_HITS' => $entr->[3]
										, '-ENTRY' => $entr->[1]
										, '-CHR' => $entr->[2]
										, '-REAL_SIZE' => $realSlength
										, '-FLANKING_FAC' => $flP
										, '-GREG_NR' =>  $doneRegs +1
										, '-FLANKING_SIZE' => $flSize
										, '-DB_NAME' =>$DBname
										, '-MIN_QN' => $minQN
										, '-RESULT_DIR' => $synResultDir
										, '-COMMAND_STR' => $cmdlineWas
										, '-ENSEMBL_DATE'=> $EnsemblD
										);
					push(@savedFiles, $csFile);
					push(@{$resRegTable->{'data'}->[-1]}, $csFile);
					
					#####$entr->[4],  $regNaame, $entr->[3], $entr->[1], $entr->[2],  $realSlength, $flP, $doneRegs +1, $flSize);
				} else {
					$outstr .= "\n";	
				}
				++$doneRegs;
				if((! $writeFlag) && ($doneRegs >= $gFl)) {
					++$breakFl;
					last;
				}
			}
			last if($breakFl);
		}
			last if($breakFl);
	      }
	
		 if($doneRegs) {
			if($restrChr) {
				print "\nchr_${restrChr}-best-${gFl}-regions:\n=============================\n" unless($silentFl);
			} else {
				print "\nglobally-best-${gFl}-regions:\n=============================\n" unless($silentFl);
			}
			print "chr\tSumOfMaxScores\tstart\trstart\tend\tQN\ttotalHSPcount\ttotalscore\t%ofRange\n" . $outstr unless($silentFl);
			
			
		} else {
			print "Nothing detected!\n" unless($silentFl);
		}
		print "global files summary:\n" . join("", @savedFiles) if(@savedFiles && (! $silentFl));
	}

 return(@savedFiles);
}








sub calcSlidingWindowStat {
	my($blastHits, $flP, $refSize, $minQN, $mScThreshold, $breadth, $stepSfrac, $silentFl)=@_;
	### blastHits are in hash-ref format of  filterBlastFile2HashEq , scoreFl defines, if the bitscore is used instead of simple counts

	$mScThreshold=0 unless(defined($mScThreshold));
	#my $breadth= $flP * $refSize;
	
	$stepSfrac = 100. unless(defined($stepSfrac));
	
	my @sortedHashKs=sort keys %{$blastHits};
	
	

	my $globalStart=SynBlast::MyUtils::max(int(getKeyPart($sortedHashKs[0], 1))-$breadth + 1, 1);
	my $globalEnd=SynBlast::MyUtils::max(int(getKeyPart($sortedHashKs[-1], 1))-$breadth + 1, 1);
	my %distribution=();
	
	
	#my %dummyC=();
	#for(my $cHitIdx=0; $cHitIdx<@sortedHashKs; ++$cHitIdx) {
	#	++$dummyC{getKeyPart($sortedHashKs[$cHitIdx], 5)};
	#}
	#return(undef) unless(scalar(keys %dummyC) >= $minQN);
	
	return(undef) unless($globalStart <= $globalEnd);
	
	my $contSkips=0; ## to skip some failing entries later on
	################################################# step-size critical
	my $stepSize=int($refSize/$stepSfrac)+1;
	
	##
	#print STDERR "\n\nGSTART= ${globalStart}, GEND= ${globalEnd}, RefSize=${refSize}, StepSize= ${stepSize}, $breadth= ${breadth}\n\n";
	##
	
	for(my $iii=$globalStart; $iii<=$globalEnd   + $stepSize; $iii+=$stepSize) {
	     my %intervalC=();  ## number of hits per query
	     #my %intervalSc=(); ## sum of scores per query
	     my %intervalMSc=(); ## keeps max-score per query
	     my $curSkips=0;
	     my $matchFl=0;
	     my $lastChitIdx=$contSkips;
	     my $realStartPos=undef;  ## contains real start pos (smallest start pos used within region $iii)
	      
	     for(my $cHitIdx=$contSkips; $cHitIdx<@sortedHashKs; ++$cHitIdx) {
		     my $currHitStartP=getKeyPart($sortedHashKs[$cHitIdx], 2);
			if( ( $currHitStartP >= $iii) && (getKeyPart($sortedHashKs[$cHitIdx], 1) <= $iii + $breadth)) { ### curBlastHit fully contained in interval?
				my $qry= getKeyPart($sortedHashKs[$cHitIdx], 5);
				my $scoore=int($blastHits->{$sortedHashKs[$cHitIdx]}->[-1]);
				
				
					#if(! exists($intervalC{$qry})) {  #no hit of same queryID already counted for current interval?
					++$distribution{$iii}->[0];  # count total sum of HSPs (hits of any queryID)
					#}
					# sum up bit-score values of hits of same interval (only considered in sum if score above a threshold)
					if($scoore >= $mScThreshold) {
						++$intervalC{$qry};					
						$distribution{$iii}->[1] += $scoore;
						#$intervalSc{$qry}+= $scoore;
						$intervalMSc{$qry} = SynBlast::MyUtils::max($scoore, $intervalMSc{$qry});
					}
				
				++$matchFl;
				push(@{$distribution{$iii}->[5]}, $blastHits->{$sortedHashKs[$cHitIdx]});
				
				$lastChitIdx=$cHitIdx;
				$realStartPos = SynBlast::MyUtils::min($currHitStartP, $realStartPos);
				
			} elsif(!  (getKeyPart($sortedHashKs[$cHitIdx], 1) < $iii + $breadth) ) { ### curBlastHit not fully contained in interval due to intervalEnd condition 
						###-> all following hits will fail too
						### so skip loop already here
				last;
			} elsif(! (getKeyPart($sortedHashKs[$cHitIdx], 2) >= $iii)) {
				### count number of continous fails from beginning (minus previously accounted skips)
				###not clear####  
				++$curSkips if(! $matchFl);
			}
			
		}
	    ## here now all values for current interval start calculated...
	    #print STDERR "slidingWindow[" . join(",", ($iii, $contSkips, $lastChitIdx, $globalStart, $globalEnd))  . "]\n";
	    if(! defined($distribution{$iii}->[1])) {
		    $distribution{$iii}->[0]=0;
		    $distribution{$iii}->[1]=0;
		    $distribution{$iii}->[2]=0;
		    $distribution{$iii}->[3]=0;
		    $distribution{$iii}->[4]=0;
		    $distribution{$iii}->[5]= [()];
		    $distribution{$iii}->[6]=0;
		    
	    } else {
		    foreach my $quuery (keys %intervalMSc) {
			    #$distribution{$iii}->[4] += (int(10 * $intervalSc{$quuery} / $intervalC{$quuery})/10.);   ## add score-per-query value to a "weighted sum of scores"
			    $distribution{$iii}->[4] += $intervalMSc{$quuery}  # now use: sum of max-scores-per-query in region
		    }
	    $distribution{$iii}->[2]= scalar(keys %intervalC); # save also the number of different queryIDs... (that have at least one representative above threshold
	    $distribution{$iii}->[3]= int(getKeyPart($sortedHashKs[$lastChitIdx], 1)); # save also the genomic end position really used of/for current interval
	    $distribution{$iii}->[6]= $realStartPos;  # save also the genomic start position really used of/for current interval
		}
	     $contSkips += int($curSkips/2.);  ## add skips from current/last session to global skipCount 
						## only half due to inexact/larger stepSizes
    	}
		
	
	return(\%distribution);
	
}


sub writeSynRegToFile {
	my($regNr, $regNaame, $regHits, $entry, $mychr, $realS, $flP, $globRegNr, $flSize, $DBname, $minQN, $synResultDir, $cmdlineWas, $EnsemblD)=
		SynBlast::MyUtils::rearrange( [qw(
									REG_NR
									REG_NAME
									REG_HITS
									ENTRY
									CHR
									REAL_SIZE
									FLANKING_FAC
									GREG_NR
									FLANKING_SIZE
									DB_NAME
									MIN_QN
									RESULT_DIR
									COMMAND_STR
									ENSEMBL_DATE
									)], @_);
	
	
	## entry is of format
	#start\trealstart\tend\tQN\ttotalHSPcount\tscore\t%ofRange
	$regNr = sprintf("%02d", $regNr);
	my $filestr = "${DBname}.${mychr}.minQN${minQN}.fac${flP}.${regNr}.blastresult";
	if(defined($globRegNr)) {
		$globRegNr = sprintf("%02d", $globRegNr);
		#$filestr =  "${DBname}.glob${globRegNr}.${mychr}.syn${minQP}.fac${flP}.${regNr}.blastresult";
		$filestr =  "${DBname}.glob${globRegNr}.${mychr}.minQN${minQN}.fac${flP}.${regNr}.blastresult";
	}
	##print "writing regions to file 
	open(FBLASTF, '>', "${synResultDir}/${filestr}")  or die "Error while creating \"${synResultDir}/${filestr}\": $!\n";
	print FBLASTF "# ${regNaame}\n"
				  . "# ${filestr}\n"
				  . "# generated by: " . $cmdlineWas . "\n"
				  ."##\n";
	print FBLASTF "#ensemblDate " . $EnsemblD 
					. "\n#synOrg " . $DBname 
					. "\n#synChr " . $mychr . "\n";
  	print FBLASTF "#synStart " . $entry->[1] 
  				. "\n#synEnd " . $entry->[2]
  				. "\n#queryCount " . $entry->[3]
				. "\n#HSPcount " . $entry->[4]
  				. "\n#maxExtension ${flP}x${realS} (" . ($flSize)  . ")"
				. "\n#PercentUsed " . $entry->[6] . "\n##\n";
	foreach my $blhiit (@{$regHits}) {
		print FBLASTF join("\t", @$blhiit) . "\n";
	}
	close FBLASTF;
	return($filestr . "\n");
}	

################################################################################
# reads all blast-data into hash of hashes (grouped by chromosome and position_data)
# optionally filtered on certain value at certain idx-position
# wublast-flag to read also wublast-tabular format
sub filterBlastFile2HashEq {
    my ($hashref, $filestr, $idx, $value, $wuformatFl) = @_;
    
    my $queryIdx = 6;
    my $targetIdx = 8;
    
    #if($wuformatFl) {
#	    $queryIdx = 6; #8;
#	    $targetIdx = 8; #10;
  #  }
    
    my $queryIdxs={()};
    $queryIdxs = $idx if(defined($idx) && (! defined($value)));  ### if only 3rd param set, this is the hash of allowed querys
    
    
    my $filename = "fhandle";
    $filename = $filestr if(! ref($filestr));
    my $myfh;
    
    eval {
    
	if($myfh=SynBlast::MyUtils::getUnixFH($filestr)) {
	
			seek($myfh,0,0); # reset postion to beginning of file.    
	}
      };
	if( $@ ) { # an error occurred
		print STDERR "ERROR (${filename}): Was not able to open files, sorry!\n";
		print STDERR "Full error is\n\n$@\n";
		exit(-1);
	}
  	
    
    
    #open(BLASTFILE, '<', "${filestr}")  or die "Error while reading \"${filestr}\": $!\n";
    #while(<BLASTFILE>) {
    while(<$myfh>) {
	if ($_ =~ m/^[^\#]/) {
	    chomp;
	    my @curHit=split; #(/\s+|\t/, $_);
	    #process curHit to get matches in chromosome orientation
	    
	    	  if($wuformatFl) {  ### 22 strd fields, optionally two more (24)
			next unless((@curHit >= 22) && (@curHit <= 24));
		  # Fields: qid   sid     E       N       Sprime  S       alignlen        nident  npos    nmism   pcident pcpos   qgaps   qgaplen sgaps   sgaplen qframe  qstart  qend    sframe  sstart  send    group   links
		   @curHit = ( $curHit[0]  # pos0: queryID
					,$curHit[1] #pos1: subjectID
					,$curHit[10]  #pos2:perc.id
				#	,$curHit[11] #pos3: perc.pos
				#	,$curHit[16] . "/" . $curHit[19] #pos4: qry/sbj-frame
					,$curHit[6]  #pos5: aln-length
					,$curHit[9]   #pos6: nmism
					,$curHit[14]  #pos7: gap opens?
					,$curHit[17]  #pos8:q.start
					,$curHit[18]   #pos9:q.end
					,$curHit[20]   #pos10:s.start
					,$curHit[21]   #pos11:s.end
					,$curHit[2]    #pos12:evalue
					,$curHit[4]     #pos13:bitscore
				);
		} else {
			next unless(@curHit < 22);
		}
		  
	    die "ERROR: tabular blast data count not 12!\n" unless(scalar(@curHit) == 12);
	    
	    next unless( (! scalar(keys %{$queryIdxs})) || exists($queryIdxs->{$curHit[0]}));   ## skip this entry if queryIdxs is defined and does not contain current queryID
	    
	    do { SynBlast::MyUtils::swapArray(\@curHit, $queryIdx, $queryIdx+1);
			SynBlast::MyUtils::swapArray(\@curHit, $targetIdx, $targetIdx+1);
		} unless ( int($curHit[$targetIdx+1]) >= int($curHit[$targetIdx]));
	  
	    
	    
	    
	    my $curH9=SynBlast::MyUtils::getLeadingZeroStr($curHit[$targetIdx+1],10);
#qx(printf "%.10d" $curHit[9]);
	    # my $curH8=qx(printf "%.10d" $curHit[8]);
	    # my $curKeyChrHits="${curH9}_${curHit[8]}_${curHit[0]}";
	    
	    my $curKeyChrHits="${curH9}_${curHit[$targetIdx]}_${curHit[$queryIdx]}_${curHit[$queryIdx +1]}_${curHit[0]}";
	    ###my $curKeyChrHits="${curH9}_${curHit[8]}_0_0_${curHit[0]}";
	    

	    #if ( (defined($value) && ("$curHit[$idx]" eq "${value}")) || (! defined($value)) ) {
	    if ( (! defined($value)) || ("$curHit[$idx]" eq "${value}") ) {
		if (defined($hashref->{$curHit[1]}->{${curKeyChrHits}})) {
		    #print STDERR "ERROR???!!!! Updating:\n";
		    #print STDERR "OldEntry= " . join(" ", @{ $hashref->{$curHit[1]}->{${curKeyChrHits}} } ) . "\n";
		    #print STDERR "NewEntry= " . join(" ", @curHit) . "\n";
		  print STDERR "E";
		}
		$hashref->{$curHit[1]}->{${curKeyChrHits}} = [ @curHit ] ;
	    }
	}
    }
    close($myfh); ###BLASTFILE);
}


sub getKeyPart {
    my ($keystr, $pos) = @_;
    if ($keystr =~ m/^(\d+)_(\d+)_(\d+)_(\d+)_(.*)$/) {
	return($1) if ($pos == 1);
	return($2) if ($pos == 2);
	return($3) if ($pos == 3);
	return($4) if ($pos == 4);
	return($5) if ($pos == 5);
    }
}








##############################################33

sub getSynRegInfo {
my ($blastFile) = @_;
open(BLASTF, '<', "${blastFile}")  or die "Error while reading ${blastFile}: $!\n";
my $counter=0;
my %infoH=();
while(<BLASTF>) {
    if ($_ =~ m/^\#/) {
	chomp;
	my @curF=split(/\s+/);
	if(@curF > 1) {
	    substr($curF[0], 0, 1) = "" if(substr($curF[0], 0, 1) eq "#");
	    $infoH{$curF[-2]}=$curF[-1];
	}
    }
}

return($infoH{'ensemblDate'},
       $infoH{'synOrg'},
       $infoH{'synChr'},
       $infoH{'synStart'},
       $infoH{'synEnd'});
}      







# reads a blastfile as-is into hash containing a list of entries grouped by idx $keyIdx
sub blastfile2Lists {
    my ($blastFile, $hashref, $keyIdx) = @_;
    defined($keyIdx) or die "Error: no grouping index selected (blastfile2Lists)\n";
    open(BLASTF, '<', "${blastFile}")  or die "Error while reading ${blastFile}: $!\n";
    my $counter=0;
    while(<BLASTF>) {
	if ($_ =~ m/^[^\#]/) {
	    chomp;
	    my @curF=split(/\s+/);
	    ($keyIdx < @curF) or die "Error: grouping index out of range (blastfile2Lists)\n";
	    push(@{$hashref->{$curF[$keyIdx]}}, [ @curF ]);
	    $counter++;
	}
    }
    close(BLASTF);
    return($counter);
}





# reads a blastfile as-is into hash containing a SORTED list of entries (hash grouped by idx $keyIdx; 
# each list sorted according to $sortIdx
# optional filtering: allowing only entries with $fvalue at $findex
sub blastfile2sortedLists {
    my ($blastFile, $hashref, $keyIdx, $sortIdx, $findex, $fvalue, $swapFl, $keyIdx2, $queryIdx, $targetIdx, $wuformatFl) = @_;
    
    defined($keyIdx) or die "Error: no grouping index selected (blastfile2sortedLists)\n";
    defined($sortIdx) or die "Error: no sort index selected (blastfile2sortedLists)\n";
    (! defined($findex)) or (defined($fvalue)) or die "Error: filter-value not defined! (blastfile2sortedLists)\n";
    if(! defined($swapFl)) { $swapFl=1; }
    
    my @blastFiles;
    if(ref $blastFile) {
      @blastFiles = @{$blastFile};
    } else {
      @blastFiles = ($blastFile);
    }
    
    $queryIdx = 6 if(! defined($queryIdx));
    $targetIdx = 8 if(! defined($targetIdx));
    
   # if($wuformatFl) {
#	    $queryIdx = 8;
#	    $targetIdx = 10;
  #  }
    
    
    my $counter=0;
    my %blastsortH=();
    my $dum=0;
    #$dum=2 if(! $swapFl);
    $dum=$targetIdx-$queryIdx if(! $swapFl);
    
    foreach $blastFile (@blastFiles) {
      open(my $BLASTF, '<', "${blastFile}")  or die "Error while reading ${blastFile}: $!\n";
      
      while(<$BLASTF>) {
	if ($_ =~ m/^[^\#]/) {
	  chomp;
	  my @curF=split; #(/\s+/);
	  next if(! @curF);
	  if($wuformatFl) {  ### 22 strd fields, optionally two more (24)
		  next unless((@curF >= 22) && (@curF <= 24));
		  # Fields: qid   sid     E       N       Sprime  S       alignlen        nident  npos    nmism   pcident pcpos   qgaps   qgaplen sgaps   sgaplen qframe  qstart  qend    sframe  sstart  send    group   links
		  next unless($curF[6]);   #pos5: aln-length should be non-zero!
		   @curF = ( $curF[0]  # pos0: queryID
					,$curF[1] #pos1: subjectID
					,$curF[10]  #pos2:perc.id
					#,$curF[11] #pos3: perc.pos
					#,$curF[16] . "/" . $curF[19] #pos4: qry/sbj-frame
					,$curF[6]  #pos5: aln-length
					,$curF[9]   #pos6: nmism
					,$curF[14]  #pos7: gap opens?
					,$curF[17]  #pos8:q.start
					,$curF[18]   #pos9:q.end
					,$curF[20]   #pos10:s.start
					,$curF[21]   #pos11:s.end
					,$curF[2]    #pos12:evalue
					,$curF[4]     #pos13:bitscore
				);
	  } else {
		  next unless(@curF < 22);
		  # Fields: query id, subject id, % identity, alignment length, mismatches, gap opens, q. start, q. end, s. start, s. end, evalue, bit score
	  }
	  next unless($curF[3]);   #pos3: aln-length should be non-zero!
	  
	  	  
	 die "ERROR: Could not read blast file properly! NCBI tabular blast output format expected (12 columns)!\n" unless(@curF < 13 );
	
	  ($sortIdx < @curF) or die "Error: sort index out of range (blastfile2sortedLists; ${sortIdx} should be less than " . @curF . ")\n";
	  (! defined($findex)) or ($findex < @curF) or die "Error: filter-index out of range (blastfile2sortedLists)\n";	    
	  
	  do { SynBlast::MyUtils::swapArray(\@curF, $queryIdx, $queryIdx+1);
	    SynBlast::MyUtils::swapArray(\@curF, $targetIdx, $targetIdx+1);
	    #my $tmp=$curHit[6]; $curHit[6]=$curHit[7]; $curHit[7]=$tmp; $tmp=$curHit[8]; $curHit[8]=$curHit[9]; $curHit[9]=$tmp;
	  } unless ( int($curF[$targetIdx+1 -$dum]) >= int($curF[$targetIdx -$dum]));
	  
	  
	  if((! defined($findex)) || ($curF[$findex] eq "${fvalue}")) {
	    push(@{$blastsortH{$curF[$sortIdx]}}, [ @curF ]);
	  }
	}
      }
      close($BLASTF);
    }
    
    ### sort according to sortIdx
    my @sortKeys= sort {$a <=> $b} (keys %blastsortH);
    for(my $li=0; $li < @sortKeys; ++$li) {
      foreach my $cEref (@{$blastsortH{$sortKeys[$li]}}) {
	($keyIdx < @{$cEref}) or die "Error: grouping index out of range (blastfile2sortedLists)\n";
	if(defined($keyIdx2)) {
	  # a second keyIdx to group on?
	  ($keyIdx2 < @{$cEref}) or die "Error: grouping index2 out of range (blastfile2sortedLists)\n";
	  push(@{$hashref->{$cEref->[$keyIdx]}->{$cEref->[$keyIdx2]}}, $cEref);
	} else {
	  push(@{$hashref->{$cEref->[$keyIdx]}}, $cEref);
	}
	$counter++;
      }
    }
    
    return($counter);
  }






sub lists2sortedLists {
    my ($inpHash, $hashref, $keyIdx, $sortIdx, $findex, $fvalue, $swapFl
        , $keyIdx2, $queryIdx, $targetIdx) = @_;
	
    defined($keyIdx) or die "Error: no grouping index selected (lists2sortedLists)\n";
    defined($sortIdx) or die "Error: no sort index selected (lists2sortedLists)\n";
    (! defined($findex)) or (defined($fvalue)) or die "Error: filter-value not defined! (lists2sortedLists)\n";
    if(! defined($swapFl)) { $swapFl=1; }
    
    
    $queryIdx = 6 if(! defined($queryIdx));
    $targetIdx = 8 if(! defined($targetIdx));
    
    my $counter=0;
    my %blastsortH=();
    my $dum=0;
    #$dum=2 if(! $swapFl);
    $dum=$targetIdx-$queryIdx if(! $swapFl);
    
    foreach my $inpKey (keys %{$inpHash}) {
	foreach my $curF (@{$inpHash->{$inpKey}}) {
	    my @curF=@{$curF};
	    ($sortIdx < @curF) or die "Error: sort index out of range (lists2sortedLists)\n";
	    (! defined($findex)) or ($findex < @curF) or die "Error: filter-index out of range (lists2sortedLists)\n";	    
	    
	    do { SynBlast::MyUtils::swapArray(\@curF, $queryIdx, $queryIdx+1);
	      SynBlast::MyUtils::swapArray(\@curF, $targetIdx, $targetIdx+1);
	  } unless ( int($curF[$targetIdx+1-$dum]) >= int($curF[$targetIdx-$dum]));

	    
	    if((! defined($findex)) || ($curF[$findex] eq "${fvalue}")) {
		push(@{$blastsortH{$curF[$sortIdx]}}, [ @curF ]);
	    }
	}
    }
    ### sort according to sortIdx
    my @sortKeys= sort {$a <=> $b} (keys %blastsortH);
    for(my $li=0; $li < @sortKeys; ++$li) {
	foreach my $cEref (@{$blastsortH{$sortKeys[$li]}}) {
	    ($keyIdx < @{$cEref}) or die "Error: grouping index out of range (lists2sortedLists)\n";
	    
	    if(defined($keyIdx2)) {
		# a second keyIdx to group on?
		($keyIdx2 < @{$cEref}) or die "Error: grouping index2 out of range (lists2sortedLists)\n";
		push(@{$hashref->{$cEref->[$keyIdx]}->{$cEref->[$keyIdx2]}}, $cEref);
		} else {
		push(@{$hashref->{$cEref->[$keyIdx]}}, $cEref);
	    }
	    $counter++;
	}
    }
    
    return($counter);
}












### creates a new InfoFile out of infoFile adding additional data columns out of addInfoHash (key is protID)
sub info2Infofile {
    my ($infoFileSrc
    , $extInfoF
    , $addInfoH
    , $verboseFl
    , $queryColName
    ) = @_;
    
    $queryColName = "protID" if(! defined($queryColName));
    
    
    ### infoFileSrc is used as source and is extended with info of addInfoHash (output to extInfoF)
#    * .protein.info  -> protein.infoW
    if(! defined($extInfoF)) {
#	$extInfoF = $infoFileSrc . "W";
	$extInfoF = $infoFileSrc
    }

    if(! (defined($addInfoH) && exists($addInfoH->{'dattype'})) ) {
	die "ERROR: No additional data available!\n";
    }
    
    $verboseFl=1 if(! defined($verboseFl));
    
    open(INFOFS, '<', "${infoFileSrc}")  or die "Error while reading ${infoFileSrc}: $!\n";
    my @newInfoCont=();

    my $headerl;
    my %infoH=();
    while(defined($headerl = <INFOFS>) && (${headerl} =~ m/^\#/)) {
    	if(${headerl} =~ m/^\#!/) {
    			chomp($headerl);
    			my @defl=split(/=/, $headerl);
    			$infoH{substr($defl[0], 2)} = $defl[-1];
    			push(@newInfoCont, $headerl . "\n") if($defl[0] ne "#!columns");
      } else {
      	#chomp($headerl);
      	push(@newInfoCont, $headerl)
      }
      
    }
    return(undef) if(! defined($headerl)); ## no content after description?
    foreach ( qw( columns focalgene organism seqregion from to )) {
    	die "Error: Could not find \"#!$_=\" definition in ${infoFileSrc}!\n" unless exists($infoH{$_});
    }
    
    #(defined($headerl = <INFOF>) && (${headerl} =~ m/^\#columns/)) or die "Error while reading ${infoFile}: 2nd line is expected to be the definition of the columns, starting with \"#columns\"\n";
    my @tcolumns = split(/,/, $infoH{"columns"});
    my %colIdx=();
    ### which idx/col is which info to find at?
    for(my $ti=0; $ti<@tcolumns; ++$ti) {
    		$colIdx{$tcolumns[$ti]}=$ti;
    }
    foreach ( $queryColName, "startPos", "endPos", "ori", "geneID", "protLength") {
    	die "Error: Could not find \"$_\" column in ${infoFileSrc}!\n" unless exists($colIdx{$_});
    }
    if(exists($colIdx{"selfScore"})) { ## selfScore already in infoFile...
    		push(@newInfoCont, "#!columns=" . join(",", @tcolumns) . "\n");
    	} else {
    		push(@newInfoCont, "#!columns=" . join(",", @tcolumns) . ",selfScore" . "\n");
    }
        
   my $colTotal=@tcolumns; #0;
   my $cPos=0;
   	do {
    	if ($headerl =~ m/^[^\#]/) {
  	    chomp($headerl);
	      my @cLi = split(/\s+/, $headerl);
	      die "ERROR: the number of data columns is not the same as in column-definition (pos${cPos})!\n" unless(@cLi eq $colTotal);
	      ++$cPos;
	 
	      if(exists($addInfoH->{$cLi[$colIdx{$queryColName}]})) {
					if(! exists($colIdx{"selfScore"})) { ## add new
						push(@newInfoCont, join("\t", @cLi, $addInfoH->{$cLi[$colIdx{$queryColName}]} ) . "\n");
		      } else { 
		    ### already data for selfComp available -> will be overwritten with new data,  if complete
		    warn("Note: already available selfComparison-score is updated/changed!\n") if($cLi[$colIdx{'selfScore'}] != $addInfoH->{$cLi[$colIdx{$queryColName}]});
		    $cLi[$colIdx{'selfScore'}] = $addInfoH->{$cLi[$colIdx{$queryColName}]};
		    push(@newInfoCont, join("\t", @cLi) . "\n");
		   }
	    } else {
		die "Could not find additional data (" . $addInfoH->{'dattype'} . ") for " . $cLi[$colIdx{$queryColName}] . "! InfoFile is not changed/created.\n";
    }
	} else {
	    push(@newInfoCont, $headerl);
	 }
 } while(defined($headerl=<INFOFS>));
    
    close(INFOFS);
    
    
    open(EINFOF, '>', "${extInfoF}")  or die "Error while creating ${extInfoF}: $!\n";
    print EINFOF join("", @newInfoCont); 
    close(EINFOF);
    print "${extInfoF} was successfully created/extended...\n" if($verboseFl);
    
    return(1);
}





sub infofile2Exonlist {
    my ($infoFile
      , $listref
      , $queryIdxHash
      , $refexonsList
      , $exonIDlist
      , $verboseFl
      , $queryColName) = @_;
    
    $verboseFl=1 if(! defined($verboseFl));
    $queryColName = "protID" if(! defined($queryColName));
    
    
    my $filename = "fhandle";
    $filename = $infoFile if(! ref($infoFile));
    my $infofh;
    
    eval {
    
	if($infofh=SynBlast::MyUtils::getUnixFH($infoFile)) {
	
			seek($infofh,0,0); # reset postion to beginning of file.    
	}
      };
	if( $@ ) { # an error occurred
		print STDERR "ERROR (${filename}): Was not able to open files, sorry!\n";
		print STDERR "Full error is\n\n$@\n";
		exit(-1);
	}
  	
	#open(INFOF, '<', "${infoFile}")  or die "Error while reading ${infoFile}: $!\n";
	  
    my $headerl;
    my %infoH=();
    while(defined($headerl = <$infofh>) && (${headerl} =~ m/^\#/)) {
    	if(${headerl} =~ m/^\#!/) {
    			chomp($headerl);
    			my @defl=split(/=/, $headerl);
    			$infoH{substr($defl[0], 2)} = $defl[-1];
      }
    }
    return(undef) if(! defined($headerl)); ## no content after description?
    foreach ( qw( columns focalgene organism seqregion from to )) {
    	die "Error: Could not find \"#!$_=\" definition in ${filename}!\n" unless exists($infoH{$_});
    }
    
    #(defined($headerl = <INFOF>) && (${headerl} =~ m/^\#columns/)) or die "Error while reading ${infoFile}: 2nd line is expected to be the definition of the columns, starting with \"#columns\"\n";
    my @tcolumns = split(/,/, $infoH{"columns"});
    my %colIdx=();
    ### which idx/col is which info to find at?
    for(my $ti=0; $ti<@tcolumns; ++$ti) {
    		$colIdx{$tcolumns[$ti]}=$ti;
    }
    foreach ( $queryColName, "startPos", "endPos", "ori", "geneID", "protLength") {
    	die "Error: Could not find \"$_\" column in ${filename}!\n" unless exists($colIdx{$_});
    }
    
    
# creating list of exons out of infoFile...
    @{$listref}=();
    %{$queryIdxHash}=() if defined($queryIdxHash);
    # stores the position of query-key in exon-list (optional)
    @{$refexonsList}=() if defined($refexonsList);
    # stores (the position of) reference-exons (IDs) for (in) exon-list (optional)
    @{$exonIDlist}=() if defined($exonIDlist);
    # stores only the Exon-IDs as list (optional)

    my $cPos=0;
    my $colTotal=@tcolumns; #0;
    my $lastRslicePos=undef;
    my $firstRslicePos=undef;
   
    do {
         # while(<INFOF>) {
			if ($headerl =~ m/^[^\#]/) {
  	    chomp($headerl);
	      my @cLi = split(/\s+/, $headerl);
	      #$colTotal = @cLi if(! $cPos); ## first line defines number of columns constraint for rest of lines
	      
			  # add estimated protein length (transcript length divided by 3) if missing protLength column????
		
	      push(@{$listref}, \@cLi ); #print ".";
	     die "ERROR: the number of data columns is not the same as in column-definition (pos${cPos})!\n" unless(@cLi eq $colTotal);
	    
	    $queryIdxHash->{$listref->[$cPos]->[$colIdx{$queryColName}]} = $cPos if defined($queryIdxHash);
	    $firstRslicePos = SynBlast::MyUtils::min($firstRslicePos, $listref->[$cPos]->[$colIdx{'startPos'}]);

	    $lastRslicePos = SynBlast::MyUtils::max($lastRslicePos, $listref->[$cPos]->[$colIdx{'endPos'}]);
	    push(@{$exonIDlist}, $listref->[$cPos]->[$colIdx{$queryColName}]) if defined($exonIDlist);
	    if(defined($refexonsList)) {
		
		if( $listref->[$cPos]->[$colIdx{'geneID'}] eq  $infoH{"focalgene"} ) {
			    # save reference exons corresponding to infofile-ensemblGeneID...
		    push(@{$refexonsList},  $listref->[$cPos]->[$colIdx{$queryColName}]);
		}
	 }
	    ++$cPos;
	}
 } while(defined($headerl=<$infofh>));
    
    close($infofh);
    my $usedFlankS = undef;
    $usedFlankS =SynBlast::MyUtils::max(          $listref->[$queryIdxHash->{$refexonsList->[0]}]->[$colIdx{'startPos'}] - $listref->[0]->[$colIdx{'startPos'}] + 1
											,   $listref->[-1]->[$colIdx{'endPos'}] - $listref->[$queryIdxHash->{$refexonsList->[-1]}]->[$colIdx{'endPos'}] + 1
										) if(@{$refexonsList});
    my $realSlength = $lastRslicePos - $firstRslicePos + 1;
    return(\%colIdx, \%infoH, $usedFlankS, $realSlength);
}



1;
